#include "core/Terminal.h"

#include <stdio.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <unistd.h>

#define ESC "\33["
#define esc(s) fputs(ESC s, stdout)

static struct termios originalTerminal;

void enterAlternativeTerminal() {
    esc("?1049h");
}

void leaveAlternativeTerminal() {
    esc("?1049l");
}

void clearTerminal() {
    esc("2J");
}

IntVector2 getTerminalSize() {
    struct winsize w;
    if(ioctl(0, TIOCGWINSZ, &w)) {
        return IV(0, 0);
    }
    return IV(w.ws_col, w.ws_row);
}

void hideCursor() {
    esc("?25l");
}

void showCursor() {
    esc("?25h");
}

void resetCursor() {
    esc("H");
}

void moveCursorLeft(int i) {
    printf(ESC "%dD", i);
}

void moveCursorRight(int i) {
    printf(ESC "%dC", i);
}

void moveCursorUp(int i) {
    printf(ESC "%dA", i);
}

void moveCursorDown(int i) {
    printf(ESC "%dB", i);
}

bool enterRawTerminal() {
    if(tcgetattr(STDIN_FILENO, &originalTerminal)) {
        return true;
    }
    struct termios raw = originalTerminal;
    raw.c_iflag &= ~(tcflag_t)(ICRNL | IXON);
    raw.c_lflag &= ~(tcflag_t)(ECHO | ICANON | IEXTEN | ISIG);
    raw.c_cc[VMIN] = 0;
    raw.c_cc[VTIME] = 0;
    return tcsetattr(STDIN_FILENO, TCSANOW, &raw);
}

bool leaveRawTerminal() {
    return tcsetattr(STDIN_FILENO, TCSAFLUSH, &originalTerminal);
}

static u8 readChar() {
    u8 c = 0;
    ssize_t bytes = read(STDIN_FILENO, &c, 1);
    return bytes <= 0 ? '\0' : c;
}

static u64 readEscapeSequence() {
    if(readChar() != '[') {
        return TERMINAL_KEY_UNKNOWN;
    }
    switch(readChar()) {
        case 'A': return TERMINAL_KEY_ARROW_UP;
        case 'B': return TERMINAL_KEY_ARROW_DOWN;
        case 'C': return TERMINAL_KEY_ARROW_RIGHT;
        case 'D': return TERMINAL_KEY_ARROW_LEFT;
        case '3':
            if(readChar() == '~') {
                return TERMINAL_KEY_ARROW_DELETE;
            }
            break;
    }
    return TERMINAL_KEY_UNKNOWN;
}

static u64 readUTF8(u64 c) {
    if((c & 0b1110'0000) == 0b1100'0000) {
        return ((c & 0b0001'1111) << 6) | (readChar() & 0b0011'1111);
    }
    // TODO: 3/4 byte UTF8 characters
    return c;
}

u64 getRawChar() {
    u8 c = readChar();
    if(c > 127) {
        return readUTF8(c);
    } else if(c == 27) {
        return readEscapeSequence();
    }
    return c;
}
